#%%
from mesa import Agent, Model
from mesa.time import RandomActivation
from mesa.batchrunner import BatchRunnerMP,BatchRunner
from mesa.datacollection import DataCollector
from math import exp, floor, log, sqrt,inf
from random import choices, uniform
import random
import numpy as np
import pandas as pd
import time
from tqdm import tqdm
import operator
import bisect
import itertools as it
import pathos
import datetime
from pymop.problem import Problem
from pymoo.algorithms.nsga2 import nsga2
from pymoo.optimize import minimize
import timeit

import warnings
warnings.filterwarnings("ignore")


###################### PARAMETERS ##########################

# Exponential Function from Cocco (2017)
# Growth function for amount of traders 
# the 1878 is for the day of the 2014.02.23


division_factor = 5000 # to make the computation more feasible
 
def grow_agent_num(t):
    
    return round(1.744e4 * exp(0.002465 * (t + 1878)) / division_factor)


# new agents calculated for each t
def new_agent_num(t):
    return grow_agent_num(t) - grow_agent_num(t-1)  



# determine agent type
def deter_agent_type(t):   
    chance_random =  percent_random * (1-chance_miner)
    chance_chartist = (1-percent_random) * (1-chance_miner)
    agent_type = ['miner', 'random','chartist']
    weights = [chance_miner, chance_random, chance_chartist]
    
    # choices picks randomly a category based on its weight, i.e. its probability
    return choices(agent_type, weights)[0]

percent_random = 0.7
chance_miner =  0.025

# Price from Blockchain.info
p_0 = 545 # BTC in US$

# Total BTC at start date: 12440250
# 60 % available and divided by 5000
b_0 = 12440250* 0.6 / division_factor

#q_init = (b_0 - np.euler_gamma)/log(grow_agent_num(0))
q_init = 231.22
q_init_cash = 20587 
q = q_init_cash * 10
total_init_hash = 28314.0905922709 / division_factor # Total Hash Rate as of 2014-02-23 in Tera Hash/s
h_init = total_init_hash/(grow_agent_num(0) * chance_miner)


# Price for one Tera Hash/s
def hash_rate_price(t):
    return (hash_rate_price_a* 10 **4) * exp(hash_rate_price_b*(t + 1272)) / 1e12
# Electricity Consumption Function in W/ Tera H/s
def electricity_consumption(t):
    return (electricity_consumption_a * 10 ** -7) * exp(electricity_consumption_b* (t + 1272)) * 1e12

elec_p = 0.07  # fixed price for KWH  https://www.ceicdata.com/en/china/electricity-price

hash_rate_price_a = 8.635
hash_rate_price_b = 0.006318

electricity_consumption_a = 4.649
electricity_consumption_b = -0.004055
# function for log normal distribution

def log_params(mu, sigma):
    return [log(mu) - 1/2 * log((sigma/mu)**2 +1), \
        sqrt(log((sigma/mu)**2 +1))]


rtrader_decide_perc = 0.1# percentage the rtrader will trade in that step

chartist_decide_perc = 0.5 # percentage the chartist will trade in that step

chartist_trade_threshhold = 0.01 # Threshhold for which above chartist makes buy order

rtrader_market_order_chance = 0.2 # chance for market order for rtradist
chartist_market_order_chance =  0.7# chance for market order for rtradist

limit_average = 1.05 # parameter for average price increase
chartist_limit_average = 1.0

chartist_dedic_wealth = log_params(0.4,0.-2)

rtrader_dedic_wealth = log_params(0.25,0.2)

rtrader_patience = log_params(3,1)

miner_next_decision_days = 60

miner_dedic_cash = log_params(0.15,0.15)

miner_dedic_btc = log_params(0.175,0.075)

chartist_time_window = 20

percent_contrary = 0.1

#### Parameters for Transaction ####

# size of transaction per day
# size of block for the whole day 1*10^6 * 6 * 24
# / size of transaction 250 


#initial fee
# 13.98013914 BTC / 68496 TXN
fee_init = 0.00020410154082 * p_0

# chance that rtrader makes transaction
rtrader_decide__trans_perc = (1/5)

trans_patience = log_params(1,0.1)

average_transaction_amount = 150

average_transaction_amount_std = 20
 
fee_std = 0.01


##### Main Model Class ###
class BlockchainModel(Model):
    
    def __init__(self, trans_limit, new_bitcoin, opt = False):
        self.running = True # for mesa batchrunner

        self.num_agents = grow_agent_num(0) # initial amount of miners

        self.schedule = RandomActivation(self)  # Scheduler from mesa

        self.curr_unique_id = 1 # used for Trader Objects

        self.t = 0 # Simulation in days
    
        self.btc_price = p_0 # Set initial BTC price equal to paramer p_0
               
        self.list_of_miner = [] 
        self.list_of_rtrader = []
        self.list_of_chartist = []

        # price development 
        self.price_list = []
       
        #fee development
        self.fee_list = [fee_init]

        self.trans_limit = round((1*10**6 * 6 * 24 / 250)/division_factor * (trans_limit/100))
        self.new_bitcoin = new_bitcoin
        
        #market lists
        self.count_markt =  0
        self.count_sell = 0
        self.count_buy = 0
        self.count_middle = 0
        self.rel_change = [0]

        # zipfl calculation for intial agents

        self.wealth_pool_zipfl_init = []

        for i in range(1, grow_agent_num(self.t) +1):
            self.wealth_pool_zipfl_init.append([(q_init/i),round((q_init_cash/i),2) ] )

        # calculated number of new agents during whole sim
        self.total_new_agents =  grow_agent_num(1500) - grow_agent_num(self.t + 1) #  39489

        # generate wealth pool for agents t > 1 
        self.wealth_pool_zipfl = []
        for i in range(1,self.total_new_agents +1):
            self.wealth_pool_zipfl.append([0,round((q/(i**(0.6))), 2)])

        random.shuffle(self.wealth_pool_zipfl)       
              
         # Create initial agents
        for i in range(self.num_agents):
            current_type = deter_agent_type(self.t)
            if current_type == 'miner':
                a = Miner(self.curr_unique_id, self,  self.wealth_pool_zipfl_init.pop(), True)
                self.list_of_miner.append(a)
            elif current_type == 'random':
                a = RTrader(self.curr_unique_id, self,  self.wealth_pool_zipfl_init.pop(), True)  
                self.list_of_rtrader.append(a)  
            elif current_type == 'chartist':
                a = Chartist(self.curr_unique_id, self,  self.wealth_pool_zipfl_init.pop(), True)  
                self.list_of_chartist.append(a)
            self.schedule.add(a)
            self.curr_unique_id = 1 + self.curr_unique_id
        

        self.tot_hash = sum([a.hash_power for a in self.list_of_miner])
        if self.tot_hash == 0:
            self.tot_hash = 0.00001
        self.hash_worth = self.tot_hash /hash_rate_price(self.t)

        self.tot_econsump = sum([a.power_consumption for a in self.list_of_miner])


        #Aggregated Data for Agents
        self.num_miner = len(self.list_of_miner)
        self.btc_miner = sum([x.btc for x in self.list_of_miner])
        self.cash_miner = sum([x.cash for x in self.list_of_miner])
        self.wealth_miner = self.cash_miner + self.btc_miner * self.btc_price + self.tot_hash/hash_rate_price(self.t)

        self.num_chartist = len(self.list_of_chartist)
        self.btc_chartist = sum([x.btc for x in self.list_of_chartist])
        self.cash_chartist = sum([x.cash for x in self.list_of_chartist])
        self.wealth_chartist = self.cash_chartist + self.btc_chartist * self.btc_price
        

        self.num_rtrader = len(self.list_of_rtrader)
        self.btc_rtrader = sum([x.btc for x in self.list_of_rtrader])
        self.cash_rtrader = sum([x.cash for x in self.list_of_rtrader])
        self.wealth_rtrader = self.cash_rtrader + self.btc_rtrader * self.btc_price
        

        # create data collector
        if opt == False: # opt flag for faster calculation in case of optimization
            self.datacollector = DataCollector(
                model_reporters={"price": 'btc_price', "last_fee":"last_fee", "hash_rate": 'tot_hash',
                'num_miner':'num_miner','num_rtrader':'num_rtrader','num_chartist':'num_chartist','btc_miner':'btc_miner',
                'btc_rtrader':'btc_rtrader','btc_chartist':'btc_chartist','cash_miner':'cash_miner',
                'cash_rtrader':'cash_rtrader','cash_chartist':'cash_chartist',
                'wealth_miner':'wealth_miner', 'wealth_chartist':'wealth_chartist', 'wealth_rtrader':'wealth_rtrader',
                'total_fees_in_btc':'total_fees_in_btc','n_txn':'n_txn','tot_econsump':'tot_econsump','hash_worth':'hash_worth'})
        else:
            self.datacollector = DataCollector(
                model_reporters={"hash_rate": 'tot_hash','wealth_miner':'wealth_miner', 'wealth_chartist':'wealth_chartist',
                 'wealth_rtrader':'wealth_rtrader'})
            
      
       
        # created market
        self.market = Market(self)

        # create transaction pool
        self.transaction_pool = TransactionPool(self)

        self.n_txn = 0

        

    # Bitcoin generation
    # until 2016-08-09 B(t) per day
    # afterewards halfed


    def bit_gen(self,t):
        if t  <= 261:
            return self.new_bitcoin/division_factor
        if t  > 261:
            return self.new_bitcoin/2 / division_factor
    
    # create agents for t > 1    
    def create_new_agents(self):
        self.new_agents = new_agent_num(self.t)
        for i in range(self.new_agents):
            current_type = deter_agent_type(self.t)
            if current_type == 'miner':
                a = Miner(self.curr_unique_id, self,  self.wealth_pool_zipfl.pop(), False)
                self.list_of_miner.append(a)
            elif current_type == 'random':
                a = RTrader(self.curr_unique_id, self,  self.wealth_pool_zipfl.pop(),  False) 
                self.list_of_rtrader.append(a)   
            elif current_type == 'chartist':
                a = Chartist(self.curr_unique_id, self,  self.wealth_pool_zipfl.pop(),  False)  
                self.list_of_chartist.append(a)
                
            self.schedule.add(a)
            self.curr_unique_id += 1

    def update_numbers(self):
        self.tot_hash = sum([a.hash_power for a in self.list_of_miner])
        if self.tot_hash == 0:
            self.tot_hash = 0.00001


        self.hash_worth = self.tot_hash /hash_rate_price(self.t)

        self.tot_btc = sum([a.btc for a in self.schedule.agents]) +\
             sum([a.residual for a in self.market.sell_list])

        self.tot_cash = sum([a.cash for a in self.schedule.agents]) + \
            sum([a.residual for a in self.market.buy_list])
        
        self.tot_econsump = sum([a.power_consumption for a in self.list_of_miner])

        self.n_txn = len(self.transaction_pool.list_of_transaction)

        self.num_miner = len(self.list_of_miner)
        self.btc_miner = sum([x.btc for x in self.list_of_miner])
        self.cash_miner = sum([x.cash for x in self.list_of_miner])
        self.wealth_miner = self.cash_miner + self.btc_miner * self.btc_price + \
            self.tot_hash/hash_rate_price(self.t) 


        self.num_chartist = len(self.list_of_chartist)
        self.btc_chartist = sum([x.btc for x in self.list_of_chartist])
        self.cash_chartist = sum([x.cash for x in self.list_of_chartist])
        self.wealth_chartist = self.cash_chartist + self.btc_chartist * self.btc_price   

        self.num_rtrader = len(self.list_of_rtrader)
        self.btc_rtrader = sum([x.btc for x in self.list_of_rtrader])
        self.cash_rtrader = sum([x.cash for x in self.list_of_rtrader])
        self.wealth_rtrader = self.cash_rtrader + self.btc_rtrader * self.btc_price


        self.last_fee = self.transaction_pool.last_fee_in_cash     
        if len(self.price_list) > 20:
            self.rel_change = (self.btc_price / self.price_list[-20]) - 1
        else:
            self.rel_change = (self.btc_price / self.price_list[-1]) - 1

        self.total_fees_in_btc = self.transaction_pool.total_fee_in_btc              
      
    def increased_activity(self):
        if len(self.price_list) >=15:
            if (self.price_list[-1]/ self.price_list[-15]) -1 > 0.016:
                decision_list = random.sample(self.list_of_miner, round(len(self.list_of_miner)/10))
                for miner_a in decision_list:
                    miner_a.makedecision()

        
    def step(self):
        self.increased_activity()
        self.create_new_agents()
        self.schedule.step()
        self.transaction_pool.execute()
        self.market.execute()
        self.market.clear_expired()
        self.price_list.append(self.btc_price)
        self.fee_list.append(self.transaction_pool.last_fee_in_cash)
        self.update_numbers()
        self.datacollector.collect(self)
        self.t = self.t +  1
      

class Market():
    def __init__(self, model):
        self.sell_list = []
        self.buy_list = []
        self.model = model
        self.trans_id = 0
       
        
    def execute(self):
        # the market sorts all sells by increasing and all buys by decreasing limits
        self.sell_list.sort(key=lambda x: (x.limit, x.order_time), reverse = False)
        self.buy_list.sort(key=lambda x: (x.limit, -x.order_time), reverse = True)

        # the loop stops when the limits do not match anymore or of the list is empty
        while self.sell_list and self.buy_list and (self.sell_list[0].limit <= self.buy_list[0].limit or  \
            self.sell_list[0].limit == 0  or  self.buy_list[0].limit == 0 ) :
            current_sell = self.sell_list[0]
            current_buy = self.buy_list[0]
                        
            self.find_price(current_sell, current_buy)
            self.maketransaction(current_sell, current_buy)


    
    # the price finding is defined by Cocco (2016)
    def find_price(self, current_sell, current_buy):
        if current_sell.limit == 0 and current_buy.limit > 0:
            self.model.btc_price = min(current_buy.limit,  self.model.btc_price)
            self.model.count_buy = self.model.count_buy + 1
        if current_sell.limit > 0 and current_buy.limit == 0:
            self.model.btc_price = max(current_sell.limit, self.model.btc_price)
            self.model.count_sell = self.model.count_sell + 1
        if current_sell.limit == 0 and current_buy.limit == 0:
            self.model.btc_price =  self.model.btc_price
            self.model.count_markt = self.model.count_markt + 1
        if current_sell.limit > 0 and current_buy.limit > 0:
            self.model.btc_price = (current_sell.limit + current_buy.limit)/2
            self.model.count_middle = self.model.count_middle + 1

    # the market will always fully execute the transaction with the smaller residual
    # and then changes the residual value of the transaction to the new residual
    def maketransaction(self, current_sell, current_buy):
        current_buyer = current_buy.agent
        current_seller = current_sell.agent
        if current_sell.residual > (current_buy.residual/ self.model.btc_price): 

            current_buyer.btc = current_buyer.btc + (current_buy.residual/ self.model.btc_price)

            current_seller.cash = round(current_seller.cash + current_buy.residual,2)
            current_sell.residual = current_sell.residual - (current_buy.residual/ self.model.btc_price)
            
            # inform miner if his total designated btc are sold
            if current_sell.upgrade:
                self.inform_miner(current_seller,current_sell.upgrade, current_buy.residual/self.model.btc_price, current_sell.trans_id)

            self.buy_list.remove(current_buy)
            

        elif current_sell.residual < (current_buy.residual/ self.model.btc_price):
            current_buyer.btc = current_buyer.btc + current_sell.residual
            current_seller.cash = round(current_seller.cash + current_sell.residual * self.model.btc_price,2)
            current_buy.residual = current_buy.residual - (current_sell.residual * self.model.btc_price)

            # inform miner if his total designated btc are sold
            if current_sell.upgrade:
                self.inform_miner(current_seller,current_sell.upgrade, current_sell.residual,\
                     current_sell.trans_id)

            self.sell_list.remove(current_sell)
            


        elif current_sell.residual == (current_buy.residual/ self.model.btc_price):
            current_buyer.btc = current_buyer.btc + current_buy.residual
            current_seller.cash = round(current_seller.cash + current_sell.residual * self.model.btc_price,2)

            # inform miner if his total designated btc are sold
            if current_sell.upgrade:
                self.inform_miner(current_seller,current_sell.upgrade, current_sell.residual, current_sell.trans_id )


            self.sell_list.remove(current_sell)
            self.buy_list.remove(current_buy)
            
    def inform_miner(self, miner,upgrade,  current_amount, trans_id):
        current_upgrade = upgrade
        current_upgrade.remain_btc = current_upgrade.remain_btc - current_amount

        current_upgrade.realized_cash = current_upgrade.realized_cash + current_amount * self.model.btc_price
        if current_upgrade.remain_btc == 0:
            current_upgrade.complete = True
      
    
    def clear_expired(self):
        # remove expired sell orders
        
        undo_list_sell = list(filter(lambda x: x.expiration == self.model.t, self.sell_list))
        self.sell_list = list(filter(lambda x: x.expiration > self.model.t, self.sell_list))
        undo_list_buy = list(filter(lambda x: x.expiration == self.model.t, self.buy_list))
        self.buy_list = list(filter(lambda x: x.expiration > self.model.t, self.buy_list))
        
        for sell in undo_list_sell:
    
            current_trader =  sell.agent
            current_trader.btc = current_trader.btc + sell.residual
        
        for buy in undo_list_buy:
       
            current_trader =  buy.agent
            current_trader.cash = round(current_trader.cash + buy.residual,2)
        
class BlockchainAgent():

    def __init__(self, unique_id, model, wealth,init):
        super().__init__()
        self.btc = wealth[0]
        self.cash = wealth[1]
        self.unique_id = unique_id
        self.model = model
        self.fee_to_pay = 0
        self.init = init
        self.new = True

    # sell and buy order are the general functions to eventually append the order to the market
    # the receive their paramters from self.create_order, self.buy_order or Miner.makedecision()
    # The orders are appended, the transactio id incremented and the wealth of the agent adjusted (the money leaves his account)
    def sellorder(self, amount, limit, patience, upgrade):
        self.model.market.sell_list.append(Order(self,upgrade, amount, limit, self.model.t, patience, \
            self.model.market.trans_id))
        self.model.market.trans_id = self.model.market.trans_id + 1
        self.btc = self.btc - amount
        #self.model.market.execute()


    def buyorder(self, amount, limit, patience, upgrade):
        self.model.market.buy_list.append(Order(self,upgrade, amount, \
            limit, self.model.t, patience, self.model.market.trans_id))
        self.model.market.trans_id = self.model.market.trans_id + 1
        self.cash = round(self.cash - amount,2)
        #self.model.market.execute()


    # In contrast to the Miner, Rtrader and Chartist don't have a decision time point, rather they have a random chance depending on 
    # the class to buy and or sell.
    def decide_to_order(self):
        if self.__class__.__name__ == 'RTrader':
            # Rtrader have 10% chance to trade with equal chance of selling and buying
            decision_p = random.uniform(0,1)
            if decision_p > rtrader_decide_perc:
                return False
            if (decision_p <= rtrader_decide_perc) and (decision_p >  rtrader_decide_perc/2 ):
                return 'Sell'
            if decision_p <   rtrader_decide_perc/2:
                return 'Buy'
            
        if self.__class__.__name__ == 'Chartist':
            # Chartist have a 50% chance to trade and make their decision whether to buy or to sell based on a historic trend, that depends on their 
            # personal time preference
            if random.uniform(0,1) < chartist_decide_perc:
                rel_change = 0

                if len(self.model.price_list) > 1:
                    if len(self.model.price_list) >= self.t_window:
                        rel_change = ((self.model.price_list[-1] / self.model.price_list[-self.t_window]) - 1)   
                       
                    elif len(self.model.price_list) < self.t_window:
                        return False
                        #rel_change = ((self.model.price_list[-1] / self.model.price_list[0]) - 1) 

              #  print(rel_change)
              

                # some chartist follow a contrary strategy and just do the opposite
                if rel_change > chartist_trade_threshhold and self.contrary == False:
                    return 'Buy'
                if rel_change > chartist_trade_threshhold and self.contrary == True:
                    return 'Sell'
                if rel_change < chartist_trade_threshhold and self.contrary == False:
                    return 'Sell'
                if rel_change < chartist_trade_threshhold and self.contrary == True:
                    return 'Buy'                
            else:
                return False

    def create_buy(self):
        # the limit price is the btc_price divided by a value from a gausian distribution
        # with mean 1 and a sigma equal to K * sigma(T_i) where T_i is the deviation of 
        # a time window of 20 of the price
        sigma_20 = max(0,np.std(list(self.model.price_list[-20:])))
        sigma_i = sigma_20 * 2.5 # 2.5 is a constant set by Cocco

        if sigma_i > 0.01:
            sigma_i = 0.01
        if sigma_i < 0.003:
            sigma_i = 0.003

        if self.__class__.__name__ == 'RTrader':
            # Rtrader make a buy order with average 25% of their cash and std of 0.2
            amount = round(self.cash * min(1,np.random.lognormal(*rtrader_dedic_wealth)),2)
            if amount > 0:
                #RTrader has 20% chance to make a markt price order

                limit = 0
                if random.uniform(0,1) >= rtrader_market_order_chance:
                    limit  = self.model.btc_price * np.random.normal( limit_average, sigma_i)                      
                # their patience is on average 3 days with one day std

                
                expiration = round(np.random.lognormal(*rtrader_patience)) + self.model.t
                self.buyorder(amount, limit, expiration,False)



        if self.__class__.__name__ == 'Chartist':
            # Chartist make a buy order with average 40% of their cash and std of 0.2
            amount = round(self.cash * min(1,np.random.lognormal(*chartist_dedic_wealth)),2)
            if amount > 0:
                #Chartist has 70% chance to make a markt price order

                limit = 0
                if random.uniform(0,1) >= chartist_market_order_chance:
                    limit  = self.model.btc_price * np.random.normal(chartist_limit_average, sigma_i)
                    
              
                expiration = self.patience + self.model.t
                self.buyorder(amount, limit, expiration,False)


    def create_sell(self):
        # the limit price is the btc_price times a value from a gausian distribution
        # with mean 1 and a sigma equal to K * sigma(T_i) where T_i is the deviation of 
        # a time window of 20 of the price
        sigma_20 = max(0,np.std(list(self.model.price_list[-20:])))
        sigma_i = sigma_20 * 2.5 # 2.5 is a constant set by Cocco           
        

        if sigma_i > 0.01:
            sigma_i = 0.01
        if sigma_i < 0.003:
            sigma_i = 0.003

        if self.__class__.__name__ == 'RTrader':
            # Rtrader make a sell order with average 25% of their cash and std of 0.2
            percent_wealth = np.random.lognormal(*rtrader_dedic_wealth)
            if percent_wealth > 1:
                percent_wealth = 1
            if percent_wealth < 0:
                percent_wealth = 0
            amount = self.btc * percent_wealth

           
            #RTrader has 70% chance to make a markt price order
            if amount > 0:
                limit = 0
                if random.uniform(0,1) >= rtrader_market_order_chance:
                    limit = self.model.btc_price / np.random.normal( limit_average, sigma_i)            

                # their patience is on average 3 days with one day std
                expiration = round(np.random.lognormal(*rtrader_patience)) + self.model.t
                self.sellorder(amount, limit, expiration,False)

        if self.__class__.__name__ == 'Chartist':
            # Chartist make a sell order with average 40% of their cash and std of 0.2
            percent_wealth = np.random.lognormal(*chartist_dedic_wealth)

            if percent_wealth > 1:
                percent_wealth = 1
            if percent_wealth < 0:
                percent_wealth = 0
            amount = self.btc * percent_wealth

                       #RTrader has 20% chance to make a markt price order
            if amount > 0 :
                limit = 0
                if random.uniform(0,1) >= chartist_market_order_chance:

                    limit = self.model.btc_price / np.random.normal(chartist_limit_average, sigma_i)            

                # their patience is always 1
                expiration =  1 + self.model.t
                self.sellorder(amount, limit, expiration,False)

    def decide_to_send(self):
        self.trans_amount = max(0.0000001, np.random.normal(average_transaction_amount/self.model.btc_price,\
             average_transaction_amount_std/self.model.btc_price))
        if self.btc * self.model.btc_price > self.trans_amount + 0.00000001 * self.model.btc_price:
            decision_p = random.uniform(0,1)
            if decision_p < rtrader_decide__trans_perc:
                return True
            else:
                return False
        else:
            return False    

    def create_transaction(self):
        amount = self.trans_amount
        recipient = random.sample([x for x in self.model.list_of_rtrader if x != self], 1)[0]
     


        if len(self.model.transaction_pool.list_of_transaction) > self.model.trans_limit:
            last_fee = self.model.transaction_pool.list_of_transaction[self.model.trans_limit].fee_in_cash 
            fee_in_cash = last_fee + 0.01 * np.random.normal(1, fee_std)

        elif len(self.model.transaction_pool.list_of_transaction) == 0:
            last_fee = self.model.transaction_pool.last_fee_in_cash
            fee_in_cash = last_fee - 0.01 * np.random.normal(1, fee_std)
        else:
            last_fee = self.model.transaction_pool.list_of_transaction[-1].fee_in_cash 
            fee_in_cash = last_fee - 0.01 * np.random.normal(1, fee_std)

        if fee_in_cash / self.model.btc_price > self.btc - amount:
            fee_in_cash = (self.btc - amount) * self.model.btc_price

        # minimum fee is one satoshi (=0.00000001 )
        if fee_in_cash <= 0.00000001 * self.model.btc_price:
            fee_in_cash = 0.00000001 * self.model.btc_price

        

        fee = fee_in_cash / self.model.btc_price
        


        curr_id = self.model.transaction_pool.trans_id
        curr_Transaction = Transaction(curr_id,self.model.t, fee, fee_in_cash, amount, self, recipient)

        bisect.insort_right(self.model.transaction_pool.list_of_transaction, curr_Transaction)
        self.model.transaction_pool.trans_id = self.model.transaction_pool.trans_id + 1
        self.btc = self.btc - amount  - fee



#### Children Classes of Blockchain Agents, the three types of Agents ###     
    
class Miner(BlockchainAgent):    
    def __init__(self, unique_id, model, wealth, init):
        super().__init__(unique_id, model, wealth, init)
        
        self.agent_type = 'Miner'
        self.list_of_hardware = []
        
         # patience for orders:
        self.patience = inf
        # upgrade list waiting for sucessful btc transaction
        self.pending_upg = []
       
        
        if self.init == True:
            # the initial Miners have a preset hardware as decided by Cocco
            w_init = electricity_consumption(0) * h_init
            self.list_of_hardware.append(Hardware(self, h_init, w_init))
            #Miner for t= 0 makes his first decision within around the next 60 days

            self.next_decision = self.model.t + round(miner_next_decision_days + np.random.normal(loc = 0, scale = 6))
            self.new = False
            self.hash_power = h_init
            self.power_consumption = w_init
        else:
            self.hash_power = 0
            self.power_consumption = 0    
            self.next_decision = None     
                    
        
        self.fee_to_pay = round(self.power_consumption/1000 * 24 * elec_p,2)
            
    
    
    # at the beginning of every t step receives a Miner bitcoins relative to his hash power compared to the total one
    def receive_bitcoin(self):
        if self.model.tot_hash > 0 :
            self.btc = self.btc + (self.model.bit_gen(self.model.t) + self.model.transaction_pool.total_fee_in_btc)\
                * (self.hash_power / self.model.tot_hash)
    
    def pay_elec_bill(self):
        self.cash = round(self.cash - self.fee_to_pay,2)
        if self.cash < 0:
            self.cash = 0
        
    
    def makedecision(self):
        
        # A Miner dedicates ~ 15% cash to buying new hardware
        gamma_cash = np.random.lognormal(*miner_dedic_cash)  #might be 0.15 or 0.3 (inconclusive in paper Cocco)
        if gamma_cash > 1:
            gamma_cash = 1     
     
              
         # A Miner dedicates ~ 17.5%  btc to buying new hardware
        gamma_btc = np.random.lognormal(*miner_dedic_btc)
        if gamma_btc > 1:
            gamma_btc = 1
     

        if self.new == True:
            self.buy_hardware( gamma_cash, gamma_btc)
        else:
            if self.btc > 0:
                # buy new hardware if constraint is fulfilled
                if electricity_consumption(self.model.t)/1000 * 24 * elec_p < self.model.btc_price * \
                    (self.model.bit_gen(self.model.t) + self.model.transaction_pool.total_fee_in_btc) * (1 / self.model.tot_hash):
                    self.buy_hardware( gamma_cash, gamma_btc)
                    
                    # if miner decides not to upgrade he sells half of the btc he was willing to spent on an upgrade
                else:
                    self.sellorder((gamma_btc * self.btc) / 2 , 0,  self.patience, False) 
            # check if hardware is outdated and sell if that's the case
            self.sell_hardware()

        self.next_decision = self.model.t + round(miner_next_decision_days + np.random.normal(loc = 0, scale = 6))

            
    def buy_hardware(self, gamma_cash, gamma_btc):
        # case for new miners for t > 0
        if self.init == False and self.new == True:
            
            
            invest = gamma_cash * self.cash
            new_hash = hash_rate_price(self.model.t) * invest
            new_elec_consumption = new_hash * electricity_consumption(self.model.t)
            self.list_of_hardware.append(Hardware(self, new_hash, new_elec_consumption ))          
            
            self.cash = round(self.cash - invest,2)
            self.update_hash_cost()
        
        # case for old miners after t > 0
        if self.new == False:                         
            
            desig_cash = gamma_cash * self.cash
            if desig_cash <= 0:
                desig_cash = 0
            desig_btc = gamma_btc * self.btc

            # Miner makes a Sell order and waits for the transaction to complete
            t_id = self.model.market.trans_id
            my_upgrade = Upgrade(t_id, desig_btc, desig_cash)
            self.sellorder(desig_btc, 0,  self.patience, my_upgrade)
            self.pending_upg.append(my_upgrade)
            self.cash = round(self.cash - desig_cash ,2)
            
            
        

    def check_pending(self):
        for upgrade in self.pending_upg:
            if upgrade.complete == True:
                new_hash = hash_rate_price(self.model.t) * (upgrade.desig_cash + (upgrade.realized_cash/2))
                if new_hash > 0:

                    new_elec_cost = new_hash * electricity_consumption(self.model.t)
                    self.list_of_hardware.append(Hardware(self, new_hash, new_elec_cost))
                    self.update_hash_cost()

                self.pending_upg.remove(upgrade)
    
    def sell_hardware(self):
        # check all units if electricity expense of unit is 20% higher than its revenue and sell hardware if that's 
        # the case
        for hardware in self.list_of_hardware:
            if hardware.power_consumption/1000 * 24 * elec_p >= 1.1 *  self.model.btc_price * \
                (self.model.bit_gen(self.model.t) + self.model.transaction_pool.total_fee_in_btc) * \
                 (hardware.hash_power / self.model.tot_hash):
                self.cash = round(self.cash + hash_rate_price(self.model.t) * hardware.hash_power,2)
                self.list_of_hardware.remove(hardware)
                self.update_hash_cost()

        
    def update_hash_cost(self):
        self.hash_power = sum([hw.hash_power for hw in self.list_of_hardware])
        self.power_consumption = sum([hw.power_consumption for hw in self.list_of_hardware])
        self.fee_to_pay = round(self.power_consumption/1000 * 24 * elec_p,2)
    
    
    def step(self):

        self.pay_elec_bill()    
        self.receive_bitcoin()
        self.check_pending()

        # if cash is zero, miners sell btc to cover electricity expenses

        if self.cash < self.fee_to_pay * 2 or self.cash < 0:
            self.sellorder(np.random.lognormal(*miner_dedic_btc) * self.btc, 0, self.patience, False)

        
        
        if (self.model.t == self.next_decision) or (self.new == True):
            self.makedecision()
        if self.new == True:
            self.new = False

class RTrader(BlockchainAgent):    
    def __init__(self, unique_id, model, wealth,init):
        super().__init__(unique_id, model, wealth,init)
        self.agent_type = 'Rtrader'
        self.hash_power = None

    
    def step(self):
        if self.new == False or (self.new == True and self.init == True):
            decision = self.decide_to_order()
            decision_trans = self.decide_to_send()
        
        
        #Rtrader always makes a buy order when he joins the market
        if self.new == True and self.init == False:
            decision = 'Buy'
            self.new = False
            decision_trans = False

        if decision != False:
            if decision == 'Sell':
                self.create_sell()
            if decision == 'Buy':
                self.create_buy()


        if decision_trans:
            self.create_transaction()
        

        if self.new == True:
            self.new = False


class Chartist(BlockchainAgent):    
    def __init__(self, unique_id, model, wealth,  init):
        super().__init__(unique_id, model, wealth, init) 
        self.patience = 1
        self.agent_type = 'Chartist'
        self.hash_power = None

        # Time window for order analysis
        self.t_window = round(np.random.normal(chartist_time_window,1))
        # variable for contrary strategy
        self.contrary = random.choices([True, False],[percent_contrary,1- percent_contrary])[0]
    def step(self):
        # 50% chance for trading that day

        if self.new == False or (self.new == True and self.init == True):
            decision = self.decide_to_order()
        if self.new == True and self.init == False:
            decision = 'Buy'
            self.new = False
            
        if decision != False:
            if decision == 'Sell':
                self.create_sell()
            if decision == 'Buy':
                self.create_buy()

        
class Hardware():
    def __init__(self, Miner, hash_power, power_consumption):
        self.hash_power = hash_power
        self.power_consumption = power_consumption


class Upgrade():
    def __init__(self, trans_id, btc, cash):
        self.trans_id = trans_id
        self.desig_btc = btc
        self.desig_cash = cash
        self.remain_btc = btc
        self.realized_cash = 0
        self.complete = False
    

class Order():
    def __init__(self, agent, upgrade, amount, limit, order_time, expiration, trans_id):
        self.agent = agent
        self.trans_id = trans_id
        self.upgrade = upgrade
        self.amount = amount
        self.residual = self.amount
        self.limit = limit
    
        self.expiration = expiration
        self.order_time = order_time



class TransactionPool():
    def __init__(self, model):
        self.model = model
        self.list_of_transaction = []
        self.last_fee_in_cash = fee_init
        self.total_fee_in_btc = 0
        self.trans_id = 0
        self.last_trans_n = round(np.random.normal(len(self.model.list_of_rtrader) * rtrader_decide__trans_perc , 5 ))
        self.trans_n_list = [self.last_trans_n,self.last_trans_n]


    def execute(self):
        #self.list_of_transaction.sort(key=lambda x: (x.fee, -x.create_time), reverse = True)
        self.last_trans_n = len(self.list_of_transaction)
        self.trans_n_list.append(len(self.list_of_transaction))
        if self.last_trans_n < self.model.trans_limit:
            self.total_fee_in_btc = sum([x.fee for x in self.list_of_transaction])
            self.last_fee_in_cash = self.list_of_transaction[-1].fee_in_cash
        else:
            self.total_fee_in_btc  = sum([x.fee for x in self.list_of_transaction[:self.model.trans_limit]])
            self.last_fee_in_cash = self.list_of_transaction[self.model.trans_limit-1].fee_in_cash


        for trans in self.list_of_transaction[:self.model.trans_limit]:
            recipient = trans.recipient
            recipient.btc = recipient.btc + trans.amount
            self.list_of_transaction.remove(trans)

        self.clear_expired()

    def clear_expired(self):
        # remove expired trans orders
        list_for_undo = list(filter(lambda x: x.expiration == self.model.t, self.list_of_transaction))
        self.list_of_transaction = list(filter(lambda x: x.expiration > self.model.t, self.list_of_transaction))

        for trans in list_for_undo:
                current_sender =  trans.sender
                current_sender.btc = current_sender.btc + trans.amount  + trans.fee 
    
    


class Transaction():
    def __init__(self, trans_id, create_time, fee, fee_in_cash, amount, sender, recipient):
        self.trans_id = trans_id
        self.expiration = round(create_time + np.random.lognormal(*trans_patience))
        self.amount = amount
        self.fee_in_cash = fee_in_cash
        self.sender = sender
        self.recipient = recipient
        self.create_time = create_time
        self.fee = fee

    def __lt__(self, others):
        return self.fee > others.fee


def run_real():
    model1 = BlockchainModel(new_bitcoin = 3600, trans_limit = 100) # translimit is measured in percent of real tlimit of 115
    for i in tqdm(range(100)):
        model1.step()
    real_run = model1.datacollector.get_model_vars_dataframe()
    real_run.to_csv('real_run.csv')




### NSGA2
def optimize_bitcoin(wealth = True, hash = False, pop_size = 20, ngen = 20, number_of_processes = 1):
    print(datetime.datetime.now())
    class MyProblem(Problem):

        def __init__(self):
            if hash == True:
                super().__init__(n_var=2, n_obj=2   , n_constr=0,
                                xl=np.array([0,0]), xu=np.array([36000,7000]))
            else: 
                super().__init__(n_var=2, n_obj=1   , n_constr=0,
                                xl=np.array([0,0]), xu=np.array([36000,7000]))
    
        def _evaluate(self, x, out, *args, **kwargs):
            f1 = []
            f2 = []
            for nsga_x in tqdm(range(len(x))):
                params = {'new_bitcoin': int(round(x[nsga_x,0])),
                "trans_limit": int(round(x[nsga_x,1])), 'opt':True}

               
                br = BatchRunnerMP(BlockchainModel,
                        iterations=25,
                        nr_processes=number_of_processes,
                        fixed_parameters=params,
                        max_steps=1000,
                display_progress = False,
                        model_reporters={'tot_hash': lambda m: m.tot_hash,'wealth_miner': lambda m: m.wealth_miner,
                        'wealth_chartist': lambda m: m.wealth_chartist,'wealth_rtrader': lambda m: m.wealth_rtrader,
                        'price':lambda m: m.btc_price, 'fee': lambda m: m.last_fee })

                br.run_all()
                run_data = br.get_model_vars_dataframe()    
                run_data = run_data.mean()
                total_wealth = run_data['wealth_miner'] + run_data['wealth_chartist'] +run_data['wealth_rtrader']
                f1.append(-total_wealth)
                f2.append(-run_data['tot_hash'])
                
            
            f1 = np.array(f1)
            f2 = np.array(f2)
            out["F"] = np.column_stack([f1,f2])

    problem = MyProblem()
 
    method = nsga2(pop_size=pop_size)


    res = minimize(problem,
                method,
                termination=('n_gen', ngen),
                seed=1,
                save_history=True,
                disp=True)
    history = []

    for gen in res.history:
        for par in gen.pop:
            current_item = [par.X[0],par.X[1],par.F[0],par.F[1],par.data['rank'],gen.n_gen]
            history.append(current_item)
    pd.DataFrame(history).to_csv('history.csv')


def runmcs(new_bitcoin, trans_limit, number_of_processes = 1):

    print(f'Bitcoin: {new_bitcoin}')
    print(f'Tlimit: {trans_limit}')
    print(datetime.datetime.now())
    tic=timeit.default_timer()

    variable_params = {'new_bitcoin': [new_bitcoin],
        "trans_limit": [trans_limit]}

    br = BatchRunnerMP(BlockchainModel,
                 iterations=25,
                 nr_processes=number_of_processes,
                 variable_parameters=variable_params,
                 max_steps=1000,
         display_progress = True,
                 model_reporters={'tot_hash': lambda m: m.tot_hash,'wealth_miner': lambda m: m.wealth_miner,
                 'wealth_chartist': lambda m: m.wealth_chartist,'wealth_rtrader': lambda m: m.wealth_rtrader,
                 'price':lambda m: m.btc_price, 'fee': lambda m: m.last_fee })

    br.run_all()
        
    run_data = br.get_model_vars_dataframe()
    run_data = run_data.mean()
    total_wealth = run_data['wealth_miner'] + run_data['wealth_chartist'] +run_data['wealth_rtrader']
    df = pd.DataFrame({'b(0)':variable_params['new_bitcoin'],'tlimit':variable_params['trans_limit'],
             'hash':run_data['tot_hash'] * 1e12, 'total_wealth': total_wealth,
                 'price':run_data['price'], 'fee':run_data['fee'],  'miner_wealth':run_data['wealth_miner'], \
                 'user_wealth':run_data['wealth_rtrader'],
                 'chartist_wealth':run_data['wealth_chartist']})   
    
    
    df.to_csv('./summary_1000_25MCS_b%s_t%s.csv' % \
            (round(new_bitcoin), round(trans_limit)))
    